function features = extractFeatures(raw, verbose)
%  Extract the following features:
%
%   - HZCRR : High Zero-Crossing rate Ratio
%   - LSTER : LowSHort Time Energy Ratio
%   - SF : Spectrum Flux
%   - BP : Band Periodicity (4 features)
%   - NFR: Noise Frame Ratio
%
    hanningWindow = hann(200).';

    windows = reshape(raw, 8000, []);

    duration = size(windows, 2);

    lster = zeros(1, duration);
    sf    = zeros(1, duration);
    bp    = zeros(4, duration);
    nfr   = zeros(1, duration); 
    hzcrr = zeros(1, duration);
    mfccv  = zeros(13,duration);

    
    Tw = 25;              % analysis frame duration (ms)
    Ts = 25;               % analysis frame shift (ms)
    alpha = 0.97;           % preemphasis coefficient
    M = 20;                 % number of filterbank channels 
    C = 12;                 % number of cepstral coefficients
    L = 22;                 % cepstral sine lifter parameter
    LF = 300;               % lower frequency limit (Hz)
    HF = 3700;              % upper frequency limit (Hz)
    fs = 8000; %muestras por segundo    
    
    
    
    
    for v = 1:duration
        if (verbose) && mod(v,500)==0,
            fprintf('%d / %d = %.2f %% \n', v, duration, v / duration * 100);
        end
        window = windows(:, v);
        frames = reshape(window, 200, 40);

        ZCRs = zeros(1, 40);
        STEs = zeros(1, 40);
        SFs  = zeros(1, 39);
        NFRs = zeros(1, 40);

        
        
        prevFFT=0;
        
        for f = 1:size(frames, 2)
            frame = frames(:, f).';
            
            ZCRs(f)  = ZCR(frame); 
            
            STEs(f) = STE(frame);
            
            NFRs(f) = NFR(frame);
            
            % SF
            wav_window = hanningWindow .* frame;
            window_length = 200;
            FFT = abs(fft(wav_window, 2*window_length));
            FFT = FFT(1 : window_length);
            if (max(FFT) ~= 0)
                FFT = FFT/max(FFT);            
            end
            %spec = fft(frame .* hanningWindow);
            if f > 1
                SFs(f-1) = SF(FFT', prevFFT');
            end
            prevFFT = FFT;

        end
        % subband energies (BP1 BP2 BP3 BP4)
        for i = 1:4
            s = subband(window, i);
            sframes = reshape(s, 200, 40);
            fbps = zeros(39,1);
            for f = 1:39
                subbandFrame = [sframes(:, f); sframes(:, f+1)];
                fbps(f) = BP(subbandFrame);
            end
            bp(i, v) = mean(fbps);
        end

                    
       MFCCs = mfcc( window(:)+eps, fs, Tw, Ts, alpha, @hamming, [LF HF], M, C+1, L );

        mfccv(:,v)= mean(MFCCs,2); 
        hzcrr(v)  = mean(sign(ZCRs - ones(1, 40) * 1.5 * mean(ZCRs))+1) / 2;
        lster(v)  = sum(sign(ones(1,40) * 0.5 * mean(STEs) - STEs)+1) / 2 / 40;
        sf(v)     = sum(SFs) / length(SFs);
        nfr(v)    = mean(NFRs < 0.32);
    end

    features = [hzcrr' lster' sf' bp' nfr' mfccv'];
    
    features(isnan(features) | isinf(features)) = 0;
end

function y = ZCR(x)
    % Zero-Crossing rate con threshold
    epsilon = 0.1;
    % Gaston version
    %y = (sum(abs(diff(x > epsilon))) + sum(abs(diff(x < -epsilon)))) / 2;
    % New version
    y = sum(abs(diff((x > epsilon) - (x < -epsilon))))/2 / length(x);
    % Original version
    %y = sum(abs(diff(x>0)))/length(x);
end

function ste = STE(audio)
% Short Time Energy
    ste = mean(audio.^2);
end

function bp = BP(s)
% Band Periodicity
% s = subband
    if size(s, 2) > 1
        s = s.';
    end
    m = length(s) / 2;
    v2 = s(m+1:end);
    normv2 = norm(v2);
    
    crossCorrelation = xcorr(v2,s);
    crossCorrelation = crossCorrelation(m+1:2*m);
    denom = conv(s.^2, ones(1,m));
    denom = denom(m:2*m-1);
    denom = flipud(sqrt(denom));
    r = crossCorrelation ./ denom / normv2;
    bp = max(r);

end

function s = subband(audio, i)
% ith subband
    % create the filterbank
    limits = [500 1000; 1000 2000; 2000 3000; 3000 3999];
    
    f = fir1(30,  limits(i,:) ./ 4000, 'bandpass'); 

    s = filter(f, 1, audio);
end

function sf = SF(frameSpectrum, prevFrameSpectrum)
% Spectrum Flux
    %sf = norm(abs(frameSpectrum)-abs(prevFrameSpectrum));
% difference spectrum (set first diff to zero)
    %sf = norm(abs(frameSpectrum-prevFrameSpectrum));
    sf = sqrt(sum((frameSpectrum - prevFrameSpectrum) .^ 2));
end

function nfr = NFR(frame)
% Noise Frame Ratio
   autocorr = xcorr(frame, 'coeff');
   autocorr(length(frame)+(-5:5)) = 0; % eliminate the central peak
   nfr = max(abs(autocorr));
end


function [ CC, FBE, frames ] = mfcc( speech, fs, Tw, Ts, alpha, window, R, M, N, L )
% MFCC Mel frequency cepstral coefficient feature extraction.
%
    % Explode samples to the range of 16 bit shorts
    if( max(abs(speech))<=1 ), speech = speech * 2^15; end;

    Nw = round( 1E-3*Tw*fs );    % frame duration (samples)
    Ns = round( 1E-3*Ts*fs );    % frame shift (samples)

    nfft = 2^nextpow2( Nw );     % length of FFT analysis 
    K = nfft/2+1;                % length of the unique part of the FFT 

    hz2mel = @( hz )( 1127*log(1+hz/700) );     % Hertz to mel function
    mel2hz = @( mel )( 700*exp(mel/1127)-700 ); % mel to Hertz function

    % Type III DCT matrix
    dctm = @( N, M )( sqrt(2.0/M) * cos( repmat([0:N-1].',1,M) ...
                                       .* repmat(pi*([1:M]-0.5)/M,N,1) ) );
    % Cepstral lifter routine
    ceplifter = @( N, L )( 1+0.5*L*sin(pi*[0:N-1]/L) );
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Preemphasis filtering
    speech = filter( [1 -alpha], 1, speech ); 

    % Framing and windowing (frames as columns)
    frames = vec2frames( speech, Nw, Ns, 'cols', window, false );

    % Magnitude spectrum computation (as column vectors)
    MAG = abs( fft(frames,nfft,1) ); 

    % Triangular filterbank with uniformly spaced filters on mel scale
    H = trifbank( M, K, R, fs, hz2mel, mel2hz ); % size of H is M x K 

    % Filterbank application to unique part of the magnitude spectrum
    FBE = H * MAG(1:K,:); 

    % DCT matrix computation
    DCT = dctm( N, M );

    % Conversion of logFBEs to cepstral coefficients through DCT
    CC =  DCT * log( FBE );

    % Cepstral lifter computation
    lifter = ceplifter( N, L );

    % Cepstral liftering gives liftered cepstral coefficients
    CC = diag( lifter ) * CC; 
    
end


function [ frames, indexes ] = vec2frames( vec, Nw, Ns, direction, window, padding )
% VEC2FRAMES Splits signal into overlapped frames using indexing.

% usage information
    usage = 'usage: [ frames, indexes ] = vec2frames( vector, frame_length, frame_shift, direction, window, padding );';

    switch( nargin )
    case { 0, 1, 2 }, error( usage );
    case 3, padding=false; window=false; direction='cols';
    case 4, padding=false; window=false; 
    case 5, padding=false; 
    end

    if( isempty(vec) || isempty(Nw) || isempty(Ns) ), error( usage ); end;
    if( min(size(vec))~=1 ), error( usage ); end;
    if( Nw==0 || Ns==0 ), error( usage ); end;

    vec = vec(:);                       % ensure column vector

    L = length( vec );                  % length of the input vector
    M = floor((L-Nw)/Ns+1);             % number of frames 


    % perform signal padding to enable exact division of signal samples into frames 
    % (note that if padding is disabled, some samples may be discarded)
    if( ~isempty(padding) )
 
        % figure out if the input vector can be divided into frames exactly
        E = (L-((M-1)*Ns+Nw));

        % see if padding is actually needed
        if( E>0 ) 

            % how much padding will be needed to complete the last frame?
            P = Nw-E;

            % pad with zeros
            if( islogical(padding) && padding ) 
                vec = [ vec; zeros(P,1) ];

            % pad with a specific numeric constant
            elseif( isnumeric(padding) && length(padding)==1 ) 
                vec = [ vec; padding*ones(P,1) ];

            % pad with a low variance white Gaussian noise
            elseif( ischar(padding) && strcmp(padding,'noise') ) 
                vec = [ vec; 1E-6*randn(P,1) ];

            % pad with a specific variance white Gaussian noise
            elseif( iscell(padding) && strcmp(padding{1},'noise') ) 
                if( length(padding)>1 ), 
                    scale = padding{2}; 
                else
                    scale = 1E-6; 
                end
                vec = [ vec; scale*randn(P,1) ];

            % if not padding required, decrement frame count
            % (not a very elegant solution)
            else
                M = M-1;

            end

            % increment the frame count
            M = M+1;
        end
    end


    % compute index matrix 
    switch( direction )

    case 'rows'                                                 % for frames as rows
        indf = Ns*[ 0:(M-1) ].';                                % indexes for frames      
        inds = [ 1:Nw ];                                        % indexes for samples
        indexes = indf(:,ones(1,Nw)) + inds(ones(M,1),:);       % combined framing indexes
    
    case 'cols'                                                 % for frames as columns
        indf = Ns*[ 0:(M-1) ];                                  % indexes for frames      
        inds = [ 1:Nw ].';                                      % indexes for samples
        indexes = indf(ones(Nw,1),:) + inds(:,ones(1,M));       % combined framing indexes
    
    otherwise
        error( sprintf('Direction: %s not supported!\n', direction) ); 

    end


    % divide the input signal into frames using indexing
    frames = vec( indexes );


    % return if custom analysis windowing was not requested
    if( isempty(window) || ( islogical(window) && ~window ) ), return; end;
    
    % if analysis window function handle was specified, generate window samples
    if( isa(window,'function_handle') )
        window = window( Nw );
    end
    
    % make sure analysis window is numeric and of correct length, otherwise return
    if( isnumeric(window) && length(window)==Nw )

        % apply analysis windowing beyond the implicit rectangular window function
        switch( direction )
        case 'rows', frames = frames * diag( window );
        case 'cols', frames = diag( window ) * frames;
        end

    end
end

function [ H, f, c ] = trifbank( M, K, R, fs, h2w, w2h )
% TRIFBANK Triangular filterbank.
%

    if( nargin~= 6 ), help trifbank; return; end; % very lite input validation

    f_min = 0;          % filter coefficients start at this frequency (Hz)
    f_low = R(1);       % lower cutoff frequency (Hz) for the filterbank 
    f_high = R(2);      % upper cutoff frequency (Hz) for the filterbank 
    f_max = 0.5*fs;     % filter coefficients end at this frequency (Hz)
    f = linspace( f_min, f_max, K ); % frequency range (Hz), size 1xK
    fw = h2w( f );

    % filter cutoff frequencies (Hz) for all filters, size 1x(M+2)
    c = w2h( h2w(f_low)+[0:M+1]*((h2w(f_high)-h2w(f_low))/(M+1)) );
    cw = h2w( c );

    H = zeros( M, K );                  % zero otherwise
    for m = 1:M 

        k = f>=c(m)&f<=c(m+1); % up-slope
        H(m,k) = (f(k)-c(m))/(c(m+1)-c(m));
        k = f>=c(m+1)&f<=c(m+2); % down-slope
        H(m,k) = (c(m+2)-f(k))/(c(m+2)-c(m+1));
       
   end
end



